/* NAME:	GReATConfigUsage
 * AUTHOR:	Attila Vizhanyo - viza@isis.vanderbilt.edu
 *
 * GReATConfig Usage Info protection & convenience proxy
 */
#ifndef GREATCONFIGUSAGE_H
#define GREATCONFIGUSAGE_H

#include <vector>
#include <map>
#include <string>
#include "GR.h"

namespace GReATConfig
{
	class UsageInformation;
	class TemporaryFileUsage;
	class FileUsage;
	class TransformationInput;
};

// Usage Info proxy class
// Provides a protected & convience way to access various information 
// in the Usage Info section of a GReATConfig model.
class GReATConfigUsage
{
public:
	// ctor; user must pass a valid 'usageInfo', no validity check performed here.
	GReATConfigUsage( GReATConfig::UsageInformation& usageInfo) :
      _usageInfo( usageInfo)
		{}
	// dtor
	virtual ~GReATConfigUsage() 
		{}
	// Returns with rule 'startRule', which is found by traversing the rules starting from 'grRootFolder'
	// as specified in the start rule path in '_usageInfo'.
	virtual void getStartRule( const GR::RootFolder& grRootFolder, GR::RewritingRule& startRule);
	// Returns temporary file usage objects
	virtual void getTemporaryFileUsages( std::vector< GReATConfig::TemporaryFileUsage>& tempFiles);
	// Returns temporary file usage, or descendant objects
	virtual void getTemporaryFileUsageKinds( std::vector< GReATConfig::TemporaryFileUsage>& tempKindFiles);
	// Returns with a list of file IDs as specified in the file usage children of '_usageInfo' 
	virtual void getFileUsageFileIDs( std::vector< std::string>& fileIDs);
	// Get various attributes belonging to 'fileID'. 
	virtual bool getMetaName( const std::string& fileID, std::string& metaName);
	virtual bool getMetaGMEParadigmName( const std::string& fileID, std::string& paradigmName);
	virtual bool getMode( const std::string& fileID, std::string& mode);
	virtual bool getRootClassName( const std::string& fileID, std::string& rootClassName);
	virtual bool getDTDFileName( const std::string& fileID, std::string& dtdFileName);
	virtual bool getRunInMemory( const std::string& fileID, bool& runInMemory);
	// Returns with transformation input children of '_usageInfo'
	virtual void getTransformationInputs( std::set< GReATConfig::TransformationInput>& inputs);
	// Transformation input object path parser.
	// TransformationInput::ObjectPath is a container of the elements of the format:
	//		ClassType[{ClassAttr1="value1"[,ClassAttr2="value2"[,...]]}]
	// Returns true if 'pathElemExpr' has valid format (parsing was successfull)
	bool parseObjectPathElem( const std::string& pathElemExpr, std::string& classType, std::vector< std::pair< std::string, std::string> >& classAttrs);
	// Returns with file usage belonging to 'fileID'. Initially builds, then search '_fileUsageMap'.
	bool getFileUsage( const std::string& fileID, GReATConfig::FileUsage& fileUsage);
	//
	void addFileUsage( const GReATConfig::TemporaryFileUsage& tfu, const std::string fileID, bool runInMem, const std::string& mode);

protected:
	// Traverse 'ruleset' recursively as specified in 'startRulePath'. 
	// Return the last rule (specified in 'startRulePath') in 'startRule'.
	void getStartRule( const std::set< GR::RewritingRule>& ruleset, std::vector< std::string>& startRulePath, GR::RewritingRule& startRule);

private:
	// Reference to usage info section of GReATConfig model.
	GReATConfig::UsageInformation& _usageInfo;
	// Keeps track of fileID - file usage associations. 
	std::map< std::string, GReATConfig::FileUsage> _fileUsageMap;
};

#endif // GREATCONFIGUSAGE_H
