/*
 *	Copyright (c) Vanderbilt University, 2001-2002
 *	ALL RIGHTS RESERVED
 *
 *	Vanderbilt University disclaims all warranties with regard to this
 *	software, including all implied warranties of merchantability
 *	and fitness.  In no event shall Vanderbilt University be liable for
 *	any special, indirect or consequential damages or any damages
 *	whatsoever resulting from loss of use, data or profits, whether
 *	in an action of contract, negligence or other tortious action,
 *	arising out of or in connection with the use or performance of
 *	this software.
 *
 *  FILE NAME:	GeneralBackend.cpp 
 *	SYNOPSIS:	General Backen for GR Engine implementation
 *  AUTHOR:		Feng Shi - feng.shi@vanderbilt.edu
 *  CREATED:	 
 *  LAST UPDATE: 
 *		8/6/04 - Feng
 *			Support the multiple datanetwork of the same meta
 *			Update the GR interface by adding a new attribute datanetworkId for ObjectWrapper
 *			Each udm object can be uniquely determined by datanetworkId and the objectId
 *
 *			Add a new map 	map<int, set<int> > deletedObjectMap;
 *			to store the deleted objects Id.
 */

#include "GeneralBackend.h"
#include "UdmBase.h"

#pragma warning ( disable : 4700 ) // Names more than 255 characters

bool GeneralBackend::GetObject(const GR::ObjectWrapper& grObj, Object& object)
{
	Udm::DataNetwork* backend = Udm::DataNetwork::GetDnById(grObj.dataNetworkID());
	if(backend!=NULL)
	{
		__int64 id = grObj.objectID();
		if(id == 0)
			return false;
		//check whether is is deleted
		map<int, set<int> >::iterator pos = deletedObjectMap.find((int)grObj.dataNetworkID());
		if(pos!=deletedObjectMap.end())
		{
			set<int> objIds = (*pos).second;
			if(objIds.find(id)!=objIds.end()) {
				return false;
			}
		}
		object=backend->ObjectById((long)id);
		if(object == Udm::null)
			return false;
	}
	else
	{
		throw udm_exception("Unknown DataNetwork.");
	}

	return true;
}

bool GeneralBackend::GetObject(int backendId, int objectId, Object& object)
{
	if(backendId==0 || objectId==0) return false;
	Udm::DataNetwork* backend = Udm::DataNetwork::GetDnById(backendId);
	if(backend!=NULL)
	{
		object=backend->ObjectById(objectId);
		if(object==Udm::null) return false;
	}
	else
		return false;
	return true;
}

bool GeneralBackend::GetGlobalObject(const string& goname, Object& object)
{
	map<string, Object>::iterator i = globalObjectMap.find(goname);
	if(i==globalObjectMap.end()) return false;
	object = (*i).second;
	return true;
}

bool GeneralBackend::GetUmlClass(const GR::ObjectWrapper& grObj, Uml::Class& clsFound)
{
/*	Udm::DataNetwork* backend  = GetBackend((string)grObj.paradigmName());
	if(backend!=NULL)
	{
		clsFound = Uml::classByName(backend->GetRootMeta(),(string)grObj.classType());
		return true;
	}
	else return false;
*/
	return GetUmlClass((string)grObj.paradigmName(),(string)grObj.nameSpace(), (string)grObj.classType(),clsFound);
}

bool GeneralBackend::GetUmlClass(const string& metaName, const string& clsName, Uml::Class& clsFound) 
{
/*	Udm::DataNetwork* backend  = GetBackend(metaName);
	if(backend!=NULL)
	{
		clsFound = Uml::classByName(backend->GetRootMeta(),clsName);
		return true;
	}
	else return false;
*/
	map<string, Uml::Diagram *>::iterator i = umlDiagramMap.find(metaName);
	if(i!=umlDiagramMap.end())
	{
//		clsFound = Uml::classByName(*((*i).second), clsName);
		clsFound = Uml::classByPath(*((*i).second), clsName);
		return true;
	}
	else return false;
}

bool GeneralBackend::GetUmlClass(const string& metaName, const string& ns_name, const string& clsName, Uml::Class& clsFound) 
{
	map<string, Uml::Diagram *>::iterator i = umlDiagramMap.find(metaName);
	if(i!=umlDiagramMap.end())
	{
		if(ns_name.empty())
			clsFound = Uml::classByName(*((*i).second), clsName);
		else
			clsFound = Uml::classByPath(*((*i).second), ns_name+"::"+clsName);
		return true;
	}
	else return false;
}

/*bool GeneralBackend::GetRootObject(const string& fileId, Object& object) 
{
	map<string, Udm::DataNetwork *>::iterator i = backendMap.find(fileId);
	if(i!=backendMap.end())
	{
		object = ((*i).second)->GetRootObject();
		return true;
	}
	else return false;
	Udm::DataNetwork* backend =	GetBackend(metaName);
	if(backend!=NULL)
	{
		object = backend->GetRootObject();
		return true;
	}
	else return false;

}*/

void GeneralBackend::GetAdjacentObjects(const GR::ObjectLink& link, Object& object, bool isSrc, set<Object>& adjObjs) {
	multiset<Object> multiadjObjs;
	GetAdjacentObjects( link, object, isSrc, multiadjObjs);
	adjObjs.insert( multiadjObjs.begin(), multiadjObjs.end());
}

string ExtractName(Udm::Object ob)
{
	Uml::Class cls= ob.type();				
	set<Uml::Attribute> attrs=cls.attributes();		
	
	// Adding parent attributes
	set<Uml::Attribute> aattrs=Uml::AncestorAttributes(cls);
	attrs.insert(aattrs.begin(),aattrs.end());

	for(set<Uml::Attribute>::iterator ai = attrs.begin();ai != attrs.end(); ai++) 
	{
		if(string(ai->type())=="String")
		{
			string str=ai->name();
			if(str=="name")
			{
				string value=ob.getStringAttr(*ai);
				if(value.empty())value="<empty string>";
				return value;
			}			
		}				
	}	
	return string("<no name specified>");
}
void GeneralBackend::GetAdjacentObjects(const GR::ObjectLink& link, Object& object, bool isSrc, multiset<Object>& multiadjObjs)
{
	set< Object> adjObjs;
	const string& linkType = link.relationshipType();
	const GR::ObjectWrapper& srcObj = link.src_end();
	const GR::ObjectWrapper& dstObj = link.dst_end();

	const string& srcType = srcObj.classType();
	const string& dstType = dstObj.classType();
	Uml::Class srcClsType, dstClsType;
	GetUmlClass(srcObj, srcClsType);
	GetUmlClass(dstObj, dstClsType);
	
	string srcName = link.srcRoleName();
	string dstName = link.dstRoleName();
	if(srcName.empty()) srcName = srcType;
	if(dstName.empty()) dstName = dstType;

	if(linkType=="Composition")    //Compositon
	{
		if(!isSrc)   // the unbound pattern object is the children
		{			
//			adjObjs=object.GetChildObjects(srcClsType); 
			if((string(link.srcRoleName())).empty() && (string(link.dstRoleName())).empty()) adjObjs=object.GetChildObjects(srcClsType); 
			else
			{
				Udm::Object::CompositionInfo cmpInfo;
				cmpInfo.strParentRoleName = link.dstRoleName();
				cmpInfo.strChildRoleName = link.srcRoleName();
				adjObjs=object.GetChildObjects(cmpInfo, srcClsType);
			}
		}
		else    // the unbound pattern object is the parent
		{
			Object parent = object.GetParent();
			if(parent) adjObjs.insert(parent);
		}
	}
	else if(linkType=="SimpleAssociation" || linkType=="ObjectMapping")
	{  
		const Uml::Class clsNull(NULL);
		Udm::Object::AssociationInfo assocInfo(clsNull);
		if(isSrc)
		{
			assocInfo.strSrcRoleName = srcName;
			assocInfo.strDstRoleName = dstName;
			adjObjs=object.GetAdjacentUniqueObjects(dstClsType,assocInfo);
		}
		else
		{
			assocInfo.strDstRoleName = srcName;
			assocInfo.strSrcRoleName = dstName;
			adjObjs=object.GetAdjacentUniqueObjects(srcClsType,assocInfo);
		}
	}	
	else if (linkType == "SimpleAssociationAny" || linkType== "ObjectMappingAny") {
		adjObjs=object.GetAdjacentUniqueObjects( dstClsType);
	}
	else if (linkType== "AssociationClassAny") {
		if(!srcObj.isAssociationClass() && !dstObj.isAssociationClass())
			throw udm_exception("None of "+(string)srcObj.name()+(string)" and "+(string)dstObj.name()+" is association class.");
		GR::ObjectWrapper peerObj= isSrc ? dstObj : srcObj;
		if( peerObj.isAssociationClass())  // Association Class Object is unbound
		{
			GetUmlClass(peerObj, dstClsType);
			string currrule= ExtractName( link.container());
			string debuginfo= string("Query assoc class objects of type: ") + ( string) dstClsType.name() + ", starting from " + ExtractName( object);
			Udm::Object::AssociationInfo ascType( dstClsType);
			adjObjs= object.GetAssociationClassObjects( NULL, ascType);//, true);
			for( std::set< Udm::Object>::const_iterator it= adjObjs.begin(); it!= adjObjs.end(); ++it) {
				assocClassAnyMap[ *it]= object;
			}
		} else //Association Class Object is bound, one of its peer ends is unbound
		{
			std::pair< Udm::Object, Udm::Object> peers= object.GetPeersFromAssociationClassObject();
			string currrule= ExtractName( link.container());
			string on, fn, sn;
			try { on= ExtractName( object);} catch ( udm_exception& e) {}
			try { fn= ExtractName( peers.first);} catch ( udm_exception& e) {}
			try { sn= ExtractName( peers.second);} catch ( udm_exception& e) {}
			string debuginfo= string("Peers of assoc class object ") + on+ " are: "+ fn + " and " + sn;
			if ( peers.first != assocClassAnyMap[ object]) {
				adjObjs.insert( peers.first);
			}
			if ( peers.second != assocClassAnyMap[ object]) {
				adjObjs.insert( peers.second);
			}
		}
	}
	else if (linkType== "AssociationClass")//AssociationClass
	{
		if(!srcObj.isAssociationClass() && !dstObj.isAssociationClass())
			throw udm_exception("None of "+(string)srcObj.name()+(string)" and "+(string)dstObj.name()+" is association class.");

		GR::ObjectWrapper peerObj;
		if(isSrc) peerObj=dstObj;
		else peerObj=srcObj;
		if(peerObj.isAssociationClass())  // Association Class Object is unbound
		{
			if(isSrc)
			{
				Udm::Object::AssociationInfo assocInfo(dstClsType);
				assocInfo.strSrcRoleName = srcName;
				assocInfo.strDstRoleName = dstName;
				adjObjs=object.GetAssociationClassObjects(NULL,assocInfo);
			}
			else
			{
				Udm::Object::AssociationInfo assocInfo(srcClsType);
				assocInfo.strSrcRoleName = dstName;
				assocInfo.strDstRoleName = srcName;
				adjObjs=object.GetAssociationClassObjects(NULL,assocInfo);
			}
		}
		else     //Association Class Object is bound, one of its peer ends is unbound
		{
			Object peerEnd;

			if(isSrc)
			{
				peerEnd = object.GetAssociationEnd(dstName);
			}
			else
			{
				peerEnd = object.GetAssociationEnd(srcName);
			}

			if(peerEnd == Udm::null) 
			{
				throw udm_exception("Object handel is Null. Please check association role names.");
			}
			else adjObjs.insert(peerEnd);
			
		}
	}
	multiadjObjs.insert( adjObjs.begin(), adjObjs.end());
}

Udm::Object GeneralBackend::CreateUdmObject(GR::ObjectWrapper &parentCO, GR::ObjectWrapper &childCO, const Udm::Object::CompositionInfo &compInfo)
{
	Object outputParent;	
	GetObject(parentCO, outputParent);
	Uml::Class childType;
	GetUmlClass(childCO, childType);

	return(outputParent.CreateObject(childType, compInfo));
}

void GeneralBackend::DeleteUdmObject(const GR::ObjectWrapper& grObj)
{
	//check whether it has been deleted
	map<int, set<int> >::iterator pos = deletedObjectMap.find((int)grObj.dataNetworkID());
	if(pos!=deletedObjectMap.end())
	{
		set<int> objIds = (*pos).second;
		if(objIds.find(grObj.objectID())!=objIds.end()) {
			return;
		}
	}

	//Check whether its parent has the "Delete" action
	set<GR::ObjectLink> links=grObj.dst();
	for(set<GR::ObjectLink>::iterator i=links.begin();i!=links.end();++i)
	{
		const GR::ObjectLink &link = *i;
		if((string)link.relationshipType()=="Composition") 
		{
			const GR::ObjectWrapper &pObj = link.dst_end();
			if((string)pObj.action()=="Delete") return;
			else break;
		}
	}
	Object udmObj;
	if(!GetObject(grObj, udmObj)) return;
//	if(!udmObj.GetParent()) return;
	DeleteUdmObject(udmObj);
}

void GeneralBackend::AddDeleteObject(Object& object)
{
	int backendId = object.__impl()->__getdn()->uniqueId();
	int objectId = object.uniqueId();

	set<int> objIds;
	map<int, set<int> >::iterator pos = deletedObjectMap.find(backendId);
	if(pos!=deletedObjectMap.end())
	{
		objIds = (*pos).second;
	}
	objIds.insert(objectId);
	deletedObjectMap[backendId] = objIds;
}

void GeneralBackend::AddDeleteObject(int backendId, int objectId)
{
	set<int> objIds;
	map<int, set<int> >::iterator pos = deletedObjectMap.find(backendId);
	if(pos!=deletedObjectMap.end())
	{
		objIds = (*pos).second;
	}

	Udm::Object deletedObj;
	if(GetObject(backendId, objectId, deletedObj))
	{
		objIds.insert(objectId);
	}
	if(!objIds.empty()) deletedObjectMap[backendId] = objIds;
}

void GeneralBackend::DeleteUdmObject(Object& object)
{
//Remove association class object
//	int backendId = 0;
//	int objectId = 0;
	set<Object> peers = object.GetAdjacentUniqueObjects();
    for(set<Object>::iterator j = peers.begin(); j!=peers.end(); j++)
	{
		Object peer = *j;
		const Uml::Class clsNull(NULL);
		Udm::Object::AssociationInfo assocInfo(clsNull);
		set<Object> assos = object.GetAssociationClassObjects(peer, assocInfo);
		for(set<Object>::iterator spos = assos.begin(); spos!=assos.end(); spos++)
		{	
			Object asso = *spos;
//			backendId = asso.__impl()->__getdn()->uniqueId();
//			objectId = asso.uniqueId();
			AddDeleteObject(asso);
			asso.DeleteObject();
	//		AddDeleteObject(backendId, objectId);
		}
	}

	AddDeleteObject(object);
//	backendId = object.__impl()->__getdn()->uniqueId();
//	objectId = object.uniqueId();
	object.DeleteObject();
//	AddDeleteObject(backendId, objectId);
}

void GeneralBackend::DeleteUdmLink(const GR::ObjectLink& grLink)
{
	const GR::ObjectWrapper& grsrc = grLink.src_end();
	const GR::ObjectWrapper& grdst = grLink.dst_end();
	Udm::Object srcObj, dstObj;
	if(GetObject(grsrc, srcObj) && GetObject(grdst, dstObj))
	{
		const Uml::Class clsNull(NULL);
		Udm::Object::AssociationInfo assoInfo(clsNull);
		assoInfo.strSrcRoleName = grLink.srcRoleName();
		assoInfo.strDstRoleName = grLink.dstRoleName();
		srcObj.DeleteLink(dstObj, assoInfo);
		//assoInfo.strDstRoleName = grLink.srcRoleName();
		//assoInfo.strSrcRoleName = grLink.dstRoleName();
		//dstObj.DeleteLink(srcObj, assoInfo);
	}
}

void GeneralBackend::CreateUdmAssociation(const GR::ObjectLink& clink, Udm::Object& srcObj, Udm::Object& dstObj)
{
	GR::ObjectWrapper srcCO = clink.src_end();
	GR::ObjectWrapper dstCO = clink.dst_end();

	string srcn = srcCO.name();
	string dstn = dstCO.name();

	string srcName = clink.srcRoleName();
	string dstName = clink.dstRoleName();
	if(srcName.empty()) srcName = srcCO.classType();
	if(dstName.empty()) dstName = dstCO.classType();

	const Uml::Class clsNull(NULL);
	Udm::Object::AssociationInfo assoInfo(clsNull);
	assoInfo.strSrcRoleName = srcName;
	assoInfo.strDstRoleName = dstName;

	if(srcObj && dstObj) 
	{	
// It needs to check here whether the link exists:
		if(!srcCO.isAssociationClass() && !dstCO.isAssociationClass())
		{
			set<Object> existedObjs;
			GetAdjacentObjects(clink, srcObj, true, existedObjs);
			if(existedObjs.find(dstObj) != existedObjs.end()) return;
		}

//		if(srcCO.isAssociationClass() || !srcObj.CreateLink(dstObj, assoInfo)) //Bug found here(4/16): the CreateLink could only work from the dst/src object, *this could not be association class object
		if(srcCO.isAssociationClass())
		{
			assoInfo.strSrcRoleName = dstName;
			assoInfo.strDstRoleName = srcName;
			if(!dstObj.CreateLink(srcObj, assoInfo))
				throw udm_exception("Create association failed with source rolename '" + dstName 
				+ "' and destination rolename '" + srcName + "'");
		}
		else //if(dstCO.isAssociationClass())
		{
			assoInfo.strSrcRoleName = srcName;
			assoInfo.strDstRoleName = dstName;
			if(!srcObj.CreateLink(dstObj, assoInfo))
				throw udm_exception("Create association failed with source rolename '" + srcName 
				+ "' and destination rolename '" + dstName + "'");
		}
	}
}

void GeneralBackend::CreateUdmAssociation(const GR::ObjectLink& clink)
{
	GR::ObjectWrapper srcCO = clink.src_end();
	GR::ObjectWrapper dstCO = clink.dst_end();

	Object srcObj, dstObj;		
	GetObject(srcCO, srcObj);
	GetObject(dstCO, dstObj);
	
	CreateUdmAssociation(clink, srcObj, dstObj);
}

void GeneralBackend::DeleteUdmAssociation(const GR::ObjectLink& clink)
{
	GR::ObjectWrapper srcCO = clink.src_end();
	GR::ObjectWrapper dstCO = clink.dst_end();

	Object srcObj, dstObj;		
	GetObject(srcCO, srcObj);
	GetObject(dstCO, dstObj);

	if((string)clink.relationshipType()=="Composition")
	{
		DeleteUdmObject(srcObj);
	}
	else
	{
		const Uml::Class clsNull(NULL);
		Udm::Object::AssociationInfo assoInfo(clsNull);
		assoInfo.strSrcRoleName = clink.srcRoleName();
		assoInfo.strDstRoleName = clink.dstRoleName();
		srcObj.DeleteLink(dstObj, assoInfo);
	}
}

void GeneralBackend::CheckNewObject(const Udm::Object& object)
{
	int backendId = object.__impl()->__getdn()->uniqueId();
	int objectId = object.uniqueId();

	set<int> objIds;
	map<int, set<int> >::iterator pos = deletedObjectMap.find(backendId);
	if(pos!=deletedObjectMap.end())
	{
		objIds = (*pos).second;
	}
	if(objIds.find(objectId)!=objIds.end()) 
	{
		objIds.erase(objectId);
		deletedObjectMap[backendId] = objIds;
	}
}
/*
void GeneralBackend::CreateObjLink(const GR::ObjectLink& clink, Object& srcObj, Object& dstObj)
{
	string srcName = clink.srcRoleName();
	string dstName = clink.dstRoleName();
	if(srcName.empty()) srcName = (GR::PatternObject::Cast(clink.src_end())).classType();
	if(dstName.empty()) dstName = (GR::PatternObject::Cast(clink.dst_end())).classType();

	const Uml::Class clsNull(NULL);
	Udm::Object::AssociationInfo assoInfo(clsNull);
	assoInfo.strSrcRoleName = srcName;
	assoInfo.strDstRoleName = dstName;

// It needs to check here whether the link exists:
	if(srcObj != &Udm::_null && dstObj != &Udm::_null) 
	{	
		bool test = srcObj.CreateLink(dstObj, assoInfo);
		if(!test)
		{
			assoInfo.strSrcRoleName = dstName;
			assoInfo.strDstRoleName = srcName;

			if(!dstObj.CreateLink(srcObj, assoInfo)) throw udm_exception("Create association fails.");
		}
		else
		{ 
			string name1, name2;
			srcObj.GetStrValue("name",name1);
			dstObj.GetStrValue("name",name2);

			GR::ObjectWrapper dobj = clink.dst_end();
		//	const string& dstType = dobj.classType();
			Uml::Class dstClsType;
			GetUmlClass(dobj, dstClsType);

			if((string)clink.relationshipType()=="AssociationClass") return;
			set<Object> adjObjs=srcObj.GetAdjacentObjects(dstClsType,assoInfo);
			if(adjObjs.find(dstObj)==adjObjs.end()) cout<<"error here"<<endl;
		}
	}
}

string GeneralBackend::getCrossObjName(Object &obj)
{					   	
	string metaname = obj.__impl()->__getdn()->GetRootMeta().name();
	return(metaname+"::"+(string)obj.type().name());
}

 void GeneralBackend::RemoveCrossLink(Object& obj1, Object& obj2)
{
	set<Object> tempSet;
	map<string, set<Object> > tempMap;

	string crname = getCrossObjName(obj2);
	map<long, map<string, set<Object> > >::iterator map_pos = crossLinks.find(obj1.uniqueId());
	if(map_pos!=crossLinks.end())
	{
		tempMap = (*map_pos).second;
		map<string, set<Object> >::iterator pos1 = tempMap.find(crname);
		if(pos1!=tempMap.end())
		{
			tempSet = (*pos1).second;
			set<Object>::iterator pos2 = tempSet.find(obj2);
			if(pos2!=tempSet.end()) tempSet.erase(pos2);
		}
	}
	
	tempMap[crname] = tempSet;
	crossLinks[obj1.uniqueId()] = tempMap;
}

void GeneralBackend::CreateCrossLink(Object& srcObj, Object& dstObj)
{
	InsertCrossLink(srcObj, dstObj);
	InsertCrossLink(dstObj, srcObj);
}

void GeneralBackend::DeleteCrossLink(Object& srcObj, Object& dstObj)
{
	RemoveCrossLink(srcObj, dstObj);
	RemoveCrossLink(dstObj, srcObj);
}

void GeneralBackend::InsertCrossLink(Object& obj1, Object& obj2)
{
	set<Object> tempSet;
	map<string, set<Object> > tempMap;

	string crname = getCrossObjName(obj2);
	map<long, map<string, set<Object> > >::iterator map_pos = crossLinks.find(obj1.uniqueId());
	if(map_pos!=crossLinks.end())
	{
		tempMap = (*map_pos).second;
		map<string, set<Object> >::iterator pos1 = tempMap.find(crname);
		if(pos1!=tempMap.end())
		{
			tempSet = (*pos1).second;
		}
		crossLinks.erase(map_pos);
	}
	tempSet.insert(obj2);
	tempMap[crname] = tempSet;
	crossLinks[obj1.uniqueId()] = tempMap;
}
*/