/*
 *	Copyright (c) Vanderbilt University, 2001-2002
 *	ALL RIGHTS RESERVED
 *
 *	Vanderbilt University disclaims all warranties with regard to this
 *	software, including all implied warranties of merchantability
 *	and fitness.  In no event shall Vanderbilt University be liable for
 *	any special, indirect or consequential damages or any damages
 *	whatsoever resulting from loss of use, data or profits, whether
 *	in an action of contract, negligence or other tortious action,
 *	arising out of or in connection with the use or performance of
 *	this software.
 *
 *  FILE NAME:	GroupEffector.cpp 
 *	SYNOPSIS:	Group Effector of Graph Rewrite Engine Implementation
 *  AUTHOR:		Feng Shi - feng.shi@vanderbilt.edu
 *  CREATED:	 08/01/06
 *  LAST UPDATE: 
 */

#pragma warning ( disable : 4786 ) // Names more than 255 characters

#include <Windows.h>
#include "sstream"
extern HINSTANCE hCINTDLL;

#include "GroupEffector.h"

#include "GReATUtils.h"
#include "RTTGenerator.h"

GroupEffector::GroupEffector(GR::RewritingRule *currRule,GeneralBackend *dataNetworkMap, vector<Match> matches) :
	boundRule(currRule),
	backendMap(dataNetworkMap),
	_matches(matches)
{
}

GroupEffector::~GroupEffector() {}

bool GroupEffector::execute()
{
	if(_matches.empty()) return true;
	const set<GR::PatternGroup> groups = boundRule->PatternGroup_kind_children();
	if(groups.empty()) return false;
	_group = *(groups.begin());	//Only one Group allowd in the Rule
	makeGroups();
	for(vector< vector<Match> >::iterator i=groupMatches.begin(); i!=groupMatches.end();++i)
	{
		vector<Match> group = *i;
		executeRightSide(group);
	}
	return true;
}

void GroupEffector::makeGroups()
{
	if(_matches.empty()) return;
	vector<Match> match_vec;
	vector<Match>::iterator mi = _matches.begin();
	match_vec.push_back(*mi);
	groupMatches.push_back(match_vec);
	mi++;
	for(; mi!=_matches.end(); ++mi)
	{
		Match the_match = *mi;
		bool found = false;
		for(vector< vector<Match> >::iterator i=groupMatches.begin(); i!=groupMatches.end();++i)
		{
			Match other_match = *((*i).begin());
			if(executeGroupRule(the_match, other_match))
			{
				found = true;
				(*i).push_back(the_match);
				break;
			}
		}
		if(!found)
		{
			vector<Match> newgroup;
			newgroup.push_back(the_match);
			groupMatches.push_back(newgroup);
		}
	}
}

void GroupEffector::executeRightSide(vector<Match> &group)
{
	map<GR::ObjectWrapper, list<Udm::Object> > boundObjs;
	for(vector<Match>::iterator gi=group.begin(); gi!=group.end();++gi)
	{
		Match match = *gi;
		for(map<GR::ObjectWrapper, Object>::const_iterator i=match.grObj2BoundObj.begin();i!=match.grObj2BoundObj.end();++i)
		{
			GR::ObjectWrapper pobj = (*i).first;
			Object uobj = (*i).second;
			if(uobj.__impl()->__getdn()->GetRootObject()!=uobj)
			{
//				if(uobj.GetParent()==Udm::null) break;
				if(uobj==Udm::null) break;
				try
				{
					if(uobj.GetParent()==Udm::null) break;
				}
				catch(udm_exception& e)
				{
					break;
				}
			}
		
			list<Udm::Object> uobjs;
			map<GR::ObjectWrapper, list<Udm::Object> >::iterator pos = boundObjs.find(pobj);		
			if(pos!=boundObjs.end())
			{
				uobjs = (*pos).second;
			}
			uobjs.push_back(uobj);
			boundObjs[pobj] = uobjs;
		}
	}	
	bindingConsequenceObject(boundObjs);
	bindingRemainingObjectLinks(boundObjs);
	executeCode(boundObjs);
	executeGroupAction(boundObjs, group);
	reqDelete(boundObjs);
	packingPassAlongPackets(boundObjs);

	//insert RTT
	createRTT(boundObjs);
}

void GroupEffector::packingPassAlongPackets(map<GR::ObjectWrapper, list<Udm::Object> > &boundObjs)
{	
	if(boundObjs.empty()) return;
	PassAlongPacket pkt;
	 map<GR::ObjectWrapper, list<Udm::Object> > updated_boundObjs;
	for(map<GR::ObjectWrapper, list<Udm::Object> >::iterator i=boundObjs.begin();i!=boundObjs.end();++i)
	{
		GR::ObjectWrapper grObj = (*i).first;

		const set<GR::PassAlong>& passAlongs = grObj.dstObjectWrapper();
		if(passAlongs.empty()) continue;

		list<Udm::Object> objlist = (*i).second;
		if(objlist.empty()) return;
		pkt[grObj] = *(objlist.begin());
		objlist.erase(objlist.begin());
		updated_boundObjs[grObj] = objlist;
	}
	_retPackets.push_back(pkt);
	packingPassAlongPackets(updated_boundObjs);
}

bool GroupEffector::bindingConsequenceObject(map<GR::ObjectWrapper, list<Udm::Object> > &boundObjs)
{
// This function need to be called recursively
	GR::ObjectWrapper parentCO;
	GR::ObjectWrapper childCO;
	
	Udm::Object::CompositionInfo compInfo;
	 
	if(getCompositionToBound(parentCO,childCO, compInfo, boundObjs))
	{
		map<GR::ObjectWrapper, list<Udm::Object> >::iterator pos = boundObjs.find(parentCO);
		if(pos==boundObjs.end())return false;
		list<Udm::Object> uObjs = (*pos).second;
		set<Udm::Object> uObjs_set;
		list2set(uObjs, uObjs_set);
		if(uObjs_set.size()>1)
		{
			throw udm_exception("Error: In one subgroup of Rule <"+(std::string)boundRule->name()+">, the size of objects bound to PatternClass <"+(std::string)parentCO.name()+"> is more than one.");
		}
		list<Udm::Object> cObjs;
		map<Udm::Object, Udm::Object> p2cObjs;
		for(list<Udm::Object>::iterator lpos = uObjs.begin(); lpos!=uObjs.end(); ++lpos)
		{
			Udm::Object pObj = *lpos;
			Udm::Object cObj;
			map<Udm::Object, Udm::Object>::iterator i = p2cObjs.find(pObj);
			if(i == p2cObjs.end())
			{
				Uml::Class childType;
				backendMap->GetUmlClass(childCO, childType);
				cObj = pObj.CreateObject(childType, compInfo);
				backendMap->CheckNewObject(cObj);
				p2cObjs[pObj] = cObj;

				//insert RTT
			}
			else
				cObj = (*i).second;
			cObjs.push_back(cObj);
		}
		boundObjs[childCO] = cObjs;

		bindingConsequenceObject(boundObjs);
	}
	return true;
}

bool GroupEffector::getCompositionToBound(GR::ObjectWrapper &parentCO,GR::ObjectWrapper &childCO, 
										  Udm::Object::CompositionInfo &compInfo,
										  map<GR::ObjectWrapper, list<Udm::Object> > &boundObjs)
{
	const set<GR::ObjectLink>& consequenceLinks=boundRule->ObjectLink_kind_children();

	for(set<GR::ObjectLink>::const_iterator i=consequenceLinks.begin();\
										   i!=consequenceLinks.end(); ++i)
	{
		const GR::ObjectLink& clink = *i;
		if((string)clink.relationshipType()!=(string)"Composition" || (string)clink.cardinality()=="0") continue;

		const GR::ObjectWrapper& dstCO = clink.dst_end();
		const GR::ObjectWrapper& srcCO = clink.src_end();

		assert((string)dstCO.type().name()==(string)srcCO.type().name());

		map<GR::ObjectWrapper, list<Udm::Object> >::iterator dpos = boundObjs.find(dstCO);
		map<GR::ObjectWrapper, list<Udm::Object> >::iterator spos = boundObjs.find(srcCO);
		
		if(dpos!=boundObjs.end() && spos==boundObjs.end())
		{
			if((string)srcCO.action()!="CreateNew" && (string)clink.action()=="CreateNew")
				throw udm_exception("Pattern object \""+(string)srcCO.name()+"\" in rule \""+(string)boundRule->name()+"\" should have action \"CreateNew\".");
	
			parentCO=dstCO;
			childCO=srcCO;
			compInfo.strParentRoleName = clink.dstRoleName();
			compInfo.strChildRoleName = clink.srcRoleName();
			return true;
		}
	}

	return false;
}

void GroupEffector::bindingRemainingObjectLinks(map<GR::ObjectWrapper, list<Udm::Object> > &boundObjs)
{
	const set<GR::ObjectLink>& consequenceLinks=boundRule->ObjectLink_kind_children();

	for(set<GR::ObjectLink>::const_iterator i=consequenceLinks.begin();\
										   i!=consequenceLinks.end(); ++i)
	{
		GR::ObjectLink clink = *i;
		if((string)clink.relationshipType()=="Composition" || (string)clink.relationshipType()=="MoveCopy")
			continue;
		if((string)clink.action()!="CreateNew")
			continue;

		list<Udm::Object> sObjs, dObjs;
		GR::ObjectWrapper srcPO = clink.src_end();
		GR::ObjectWrapper dstPO = clink.dst_end();
		if(boundObjs.find(srcPO) == boundObjs.end() || boundObjs.find(dstPO) == boundObjs.end())
		{
			throw udm_exception("Group Error.");
		}
		sObjs = boundObjs[srcPO];
		dObjs = boundObjs[dstPO];

		list<Udm::Object>::const_iterator site = sObjs.begin();
		list<Udm::Object>::const_iterator dite = dObjs.begin();
		
		while(site!=sObjs.end() && dite!=dObjs.end())
		{
			Udm::Object sObj = *site;
			Udm::Object dObj = *dite;
			backendMap->CreateUdmAssociation(clink, sObj, dObj);
			site++;
			dite++;
		}
	}
}

bool GroupEffector::GetCopyPeer(Udm::Object &fromObj, Udm::Object &toObj, Udm::Object &fromParent, Udm::Object &copy_Parent, 
								map<pair<Udm::Object, Udm::Object>, UdmUtil::copy_assoc_map>& copyMap)
{
	pair<Udm::Object, Udm::Object> c_p;
	c_p.first = fromParent;
	c_p.second = copy_Parent;
	map<pair<Udm::Object, Udm::Object>, UdmUtil::copy_assoc_map>::iterator pos = copyMap.find(c_p);
	if(pos == copyMap.end())
	{
		//if rootobj, then false;
		if(fromParent.__impl()->__getdn()->GetRootObject()==fromParent) return false;
		Udm::Object fromObj_parent = fromParent.GetParent();
		return GetCopyPeer(fromObj,toObj,fromObj_parent,copy_Parent,copyMap);
	}
	else
	{
		UdmUtil::copy_assoc_map copy_peers_map = (*pos).second;
		if(copy_peers_map.find(fromObj)==copy_peers_map.end()) return false;
		toObj = copy_peers_map[fromObj];
		return true;
	}
}

void GroupEffector::makeInstCopyAssocMap(Udm::Object &from_obj, Udm::Object &to_obj, UdmUtil::copy_assoc_map &copy_map)
{
	copy_map[from_obj] = to_obj;	
	to_obj.__impl()->CopyAttributesFrom(from_obj.__impl());   //This line is for deep copy the attributes
	set<Udm::Object> from_children = from_obj.GetChildObjects();
	set<Udm::Object> to_children = to_obj.GetChildObjects();
	for(set<Udm::Object>::iterator i=from_children.begin();i!=from_children.end();++i)
	{
		Udm::Object from_child = *i;
		for(set<Udm::Object>::iterator j=to_children.begin();j!=to_children.end();++j)
		{
			Udm::Object to_child = *j;
			if(from_child.archetype() == to_child.archetype())
			{
				makeInstCopyAssocMap(from_child, to_child, copy_map);
				to_children.erase(to_child);
				break;
			}
		}
	}
}

void GroupEffector::executeGroupAction(map<GR::ObjectWrapper, list<Udm::Object> > &boundObjs, vector<Match> &group)
{
	if((string)_group.action()=="Bound")
		return;

	set<GR::GRBase> grObjs = _group.members();  //get group members
	set<Udm::Object> groupObjs;                 //objects in group
	set<Udm::Object> groupAssocs;                //association objects in group
	set<GR::ObjectWrapper> grAssocs;            //GR assocition objectwrapper as member of the group
	set<GR::ObjectLink> groupLinks;

	map<Udm::Object, pair<Udm::Object, Udm::Object> > assoc_with_peer_ends; 
	map<Udm::Object, pair<string, string> > assoc_with_rolenames;
	map<Udm::Object, set<Udm::Object> > set_with_members;
	map<Udm::Object, pair<string, string> > set_with_rolenames;

	for(set<GR::GRBase>::const_iterator i=grObjs.begin();i!=grObjs.end();++i)
	{
		GR::GRBase grObj = *i;
		if((string)grObj.type().name()=="PatternObject")
		{
			GR::PatternObject pObj = GR::PatternObject::Cast(grObj);
			list<Udm::Object> bobjs = boundObjs[pObj];

			Uml::Class clsType;
			backendMap->GetUmlClass(pObj, clsType);

			for(list<Udm::Object>::iterator li=bobjs.begin();li!=bobjs.end();++li)
			{
				if(pObj.isAssociationClass() && (string)clsType.stereotype()=="Connection")
				{
					groupAssocs.insert(*li);
					ProcessGRAssociation(group, pObj, assoc_with_peer_ends, assoc_with_rolenames);
				}
				else
                    groupObjs.insert(*li);
			}
		}
		else //ObjectLink
		{
			groupLinks.insert(GR::ObjectLink::Cast(grObj));
			ProcessGRLink(group, GR::ObjectLink::Cast(grObj), assoc_with_peer_ends, assoc_with_rolenames, set_with_members, set_with_rolenames);
		}
	}
	
	if((string)_group.action()=="Copy")
	{
		CopyGroup(boundObjs,groupObjs,assoc_with_peer_ends,assoc_with_rolenames,set_with_members,set_with_rolenames);
	}
	else if((string)_group.action()=="Move")
	{
		MoveGroup(boundObjs,groupObjs,assoc_with_peer_ends,assoc_with_rolenames,set_with_members,set_with_rolenames);
	}
	else//"Delete"
	{
		DeleteGroup(group, groupLinks, groupAssocs, groupObjs);
	}
}

void GroupEffector::ProcessGRLink(vector<Match> &group, GR::ObjectLink &grlink,
								   map<Udm::Object, pair<Udm::Object, Udm::Object> > &assoc_with_peer_end,
								   map<Udm::Object, pair<string, string> > &assoc_with_rolenames, 
								   map<Udm::Object, set<Udm::Object> > &set_with_members,
								   map<Udm::Object, pair<string, string> > &set_with_rolenames)
{
	GR::ObjectWrapper src_grObj = grlink.src_end();
	GR::ObjectWrapper dst_grObj = grlink.dst_end();
	
	string srcRoleName = grlink.srcRoleName();
	string dstRoleName = grlink.dstRoleName();
	
	Uml::Class clsDstType;
	backendMap->GetUmlClass(dst_grObj, clsDstType);

	if((string)clsDstType.stereotype()=="Set")
	{
		pair<string, string> rolenames;
		rolenames.first = grlink.srcRoleName();
		rolenames.second = grlink.dstRoleName();
		for(vector<Match>::iterator j=group.begin();j!=group.end();++j)
		{
			Match match = *j;
			Udm::Object memberObj = match.grObj2BoundObj[src_grObj];
			Udm::Object setObj = match.grObj2BoundObj[dst_grObj];
			set<Udm::Object> members;
			map<Udm::Object, set<Udm::Object> >::iterator setPos = set_with_members.find(setObj);
			if(setPos!=set_with_members.end())
				members = (*setPos).second;
			members.insert(memberObj);
			set_with_members[setObj] = members;
			set_with_rolenames[setObj] = make_pair(srcRoleName, dstRoleName);
		}
	}
	else
	{
		const Uml::Class clsNull(NULL);
		Udm::Object::AssociationInfo assoInfo(clsNull);
		assoInfo.strSrcRoleName = srcRoleName;
		assoInfo.strDstRoleName = dstRoleName;

		for(vector<Match>::iterator j=group.begin();j!=group.end();++j)
		{
			Match match = *j;
			Udm::Object src_obj = match.grObj2BoundObj[src_grObj];
			Udm::Object dst_obj = match.grObj2BoundObj[dst_grObj];
			multiset< pair<Udm::Object, Udm::Object> > assoc_with_peerends = src_obj.GetAdjacentObjectsWithAssocClasses(clsDstType, assoInfo);  
			if(assoc_with_peerends.empty()) return;
			for(multiset< pair<Udm::Object, Udm::Object> >::iterator pos=assoc_with_peerends.begin();pos!=assoc_with_peerends.end();++pos)
			{
				pair<Udm::Object, Udm::Object> assoc_w_peerend = *pos;
				if(assoc_w_peerend.first == dst_obj)
				{
					assoc_with_peer_end[assoc_w_peerend.second]=make_pair(src_obj,dst_obj);
					assoc_with_rolenames[assoc_w_peerend.second]=make_pair(srcRoleName, dstRoleName);
					break;
				}
			}
		}
	}
}


void GroupEffector::ProcessGRAssociation(vector<Match> &group, GR::PatternObject &grAssoc,
								   map<Udm::Object, pair<Udm::Object, Udm::Object> > &assoc_with_peer_end,
								   map<Udm::Object, pair<string, string> > &assoc_with_rolenames) 
{
	set<GR::ObjectLink> srclinks = grAssoc.src();
	GR::ObjectLink srclink = *(srclinks.begin());
	GR::ObjectWrapper srcObjWrapper = srclink.src_end();

	set<GR::ObjectLink> dstlinks = grAssoc.dst();
	GR::ObjectLink dstlink;
	for(set<GR::ObjectLink>::iterator dlinks_it=dstlinks.begin();dlinks_it!=dstlinks.end();++dlinks_it)
	{
		dstlink = *dlinks_it;
		if((string)dstlink.name()=="Dst") break;
	}
	GR::ObjectWrapper dstObjWrapper = dstlink.dst_end();

	string srcRoleName = srclink.srcRoleName();
	string dstRoleName = srclink.dstRoleName();

	for(vector<Match>::iterator i=group.begin();i!=group.end();++i)
	{
		Match match = *i;
		Udm::Object assocObj = match.grObj2BoundObj[grAssoc];
		Udm::Object srcObj = match.grObj2BoundObj[srcObjWrapper];		
		Udm::Object dstObj = match.grObj2BoundObj[dstObjWrapper];
		assoc_with_rolenames[assocObj] = make_pair(srcRoleName, dstRoleName);
		assoc_with_peer_end[assocObj] = make_pair(srcObj, dstObj);
	}
}


void GroupEffector::CopyGroup(map<GR::ObjectWrapper, list<Udm::Object> > &boundObjs, set<Udm::Object> &groupObjs,
							  map<Udm::Object, pair<Udm::Object, Udm::Object> > &assoc_with_peer_end,
							  map<Udm::Object, pair<string, string> > &assoc_with_rolenames, 
							  map<Udm::Object, set<Udm::Object> > &set_with_members,
							  map<Udm::Object, pair<string, string> > &set_with_rolenames)
{
	//Get all copy destination objects
	set<GR::ObjectLink> nps = _group.dst();
	if(nps.empty())
		throw udm_exception("In the rule: "+(string)boundRule->name()+" the Group has no copy destination.\n");
	set<Udm::Object> copy_parents;
	for(set<GR::ObjectLink>::iterator i=nps.begin();i!=nps.end();++i)
	{
		GR::ObjectLink np = *i;
		GR::ObjectWrapper mp = np.dst_end();
		list<Udm::Object> uobjs = boundObjs[mp];
		for(list<Udm::Object>::iterator li=uobjs.begin();li!=uobjs.end();++li)
		{
			copy_parents.insert(*li);
		}
	}

	map<pair<Udm::Object, Udm::Object>, UdmUtil::copy_assoc_map> copyMap;  //pair<to_be_copied_Obj, parent_of_copy>
	set<Udm::Object>::iterator pit;
	for(set<Udm::Object>::iterator oi=groupObjs.begin(); oi!=groupObjs.end();++oi)
	{
		Udm::Object from_obj = *oi;
		
		bool isInstanceObj = from_obj.isInstance() || from_obj.isSubtype();
		Udm::Object from_arch_obj;
		if(isInstanceObj)
			from_arch_obj = from_obj.archetype();

		Uml::Class childType = from_obj.type();
		
		for(pit=copy_parents.begin();pit!=copy_parents.end();++pit)
		{
			Udm::Object copy_parent = *pit;				
			Udm::Object to_obj;
			UdmUtil::copy_assoc_map copy_map;
			if(isInstanceObj)
			{
				//CreateInstance
				to_obj = copy_parent.CreateObject(childType,from_arch_obj,from_obj.isSubtype());
				backendMap->CheckNewObject(to_obj);
//				to_obj.__impl()->CopyAttributesFrom(from_obj.__impl());
				makeInstCopyAssocMap(from_obj, to_obj, copy_map);
			}
			else
			{
				to_obj = copy_parent.CreateObject(childType);
				backendMap->CheckNewObject(to_obj);
	//			UdmUtil::CopyOpts opts = { false, false };
				UdmUtil::CopyObjectHierarchy(from_obj.__impl(), to_obj.__impl(), to_obj.__impl()->__getdn(), copy_map);
				copy_map[from_obj] = to_obj;
			}
			copyMap[make_pair(from_obj, copy_parent)] = copy_map;
		}
	}

	//Create the association objects and reset the association relationship
	for(map<Udm::Object, pair<Udm::Object, Udm::Object> >::iterator ai=assoc_with_peer_end.begin();ai!=assoc_with_peer_end.end();++ai)
	{
		Udm::Object from_assocObj = (*ai).first;
		Uml::Class childType = from_assocObj.type();
		Udm::Object::CompositionInfo compInfo;

		Udm::Object from_srcObj = ((*ai).second).first;
		Udm::Object from_dstObj = ((*ai).second).second;
		
		pair<string, string> rolenames = assoc_with_rolenames[from_assocObj];
		const Uml::Class clsNull(NULL);
		Udm::Object::AssociationInfo src_assoInfo(clsNull);
		src_assoInfo.strSrcRoleName = rolenames.first;
		src_assoInfo.strDstRoleName = rolenames.second;

		Udm::Object::AssociationInfo dst_assoInfo(clsNull);
		dst_assoInfo.strSrcRoleName = rolenames.second;
		dst_assoInfo.strDstRoleName = rolenames.first;
		
		for(pit=copy_parents.begin();pit!=copy_parents.end();++pit)
		{
			Udm::Object copy_parent = *pit;				

			//Reset the association
			Udm::Object to_srcObj, to_dstObj;
			if(GetCopyPeer(from_srcObj,to_srcObj,from_srcObj,copy_parent,copyMap) &&
				GetCopyPeer(from_dstObj,to_dstObj,from_dstObj,copy_parent,copyMap) )
			{
				Udm::Object to_assocObj = copy_parent.CreateObject(childType, compInfo);
				backendMap->CheckNewObject(to_assocObj);
				UdmUtil::copy_assoc_map copy_map;
	//			UdmUtil::CopyOpts opts = { false, false };
				UdmUtil::CopyObjectHierarchy(from_assocObj.__impl(), to_assocObj.__impl(), to_assocObj.__impl()->__getdn(), copy_map);
				copyMap[make_pair(from_assocObj, copy_parent)] = copy_map;			

				to_srcObj.CreateLink(to_assocObj, src_assoInfo);
				to_dstObj.CreateLink(to_assocObj, dst_assoInfo);
			}
		}
	}

	//reset the set-member relationship
	for(map<Udm::Object, set<Udm::Object> >::iterator si=set_with_members.begin();si!=set_with_members.end();++si)
	{
		Udm::Object from_set = (*si).first;
		set<Udm::Object> from_members = (*si).second;

		pair<string, string> rolenames = set_with_rolenames[from_set];
		const Uml::Class clsNull(NULL);
		Udm::Object::AssociationInfo set_assoInfo(clsNull);
		set_assoInfo.strSrcRoleName = rolenames.first;
		set_assoInfo.strDstRoleName = rolenames.second;

		for(pit=copy_parents.begin();pit!=copy_parents.end();++pit)
		{
			Udm::Object copy_parent = *pit;				
			Udm::Object to_set;
			if(GetCopyPeer(from_set,to_set,from_set,copy_parent,copyMap))
			{
				for(set<Udm::Object>::iterator mi=from_members.begin();mi!=from_members.end();++mi)
				{
					Udm::Object from_member = *mi;
					Udm::Object to_member;
					if(GetCopyPeer(from_member,to_member,from_member,copy_parent,copyMap))
					{
						to_member.CreateLink(to_set, set_assoInfo);
					}
				}
			}		
		}
	}	
}

void GroupEffector::MoveGroup(map<GR::ObjectWrapper, list<Udm::Object> > &boundObjs, set<Udm::Object> &groupObjs,
							  map<Udm::Object, pair<Udm::Object, Udm::Object> > &assoc_with_peer_end,
							  map<Udm::Object, pair<string, string> > &assoc_with_rolenames, 
							  map<Udm::Object, set<Udm::Object> > &set_with_members,
							  map<Udm::Object, pair<string, string> > &set_with_rolenames)
{
	//Get the destination of to-be-moved group
	set<GR::ObjectLink> nps = _group.dst();
	if(nps.empty())
		throw udm_exception("In the rule: "+(string)boundRule->name()+" the Group has no move destination.\n");
	if(nps.size()>1)
		throw udm_exception("In the rule: "+(string)boundRule->name()+" the Group has more than one move destinations.\n");
	GR::ObjectLink np = *(nps.begin());
	GR::ObjectWrapper mp = np.dst_end();
	list<Udm::Object> moveDsts_list = boundObjs[mp];
	Udm::Object moveDst = *(moveDsts_list.begin());

	//First create a new association objects in the destination
	map<Udm::Object, pair<Udm::Object, Udm::Object> > newAssocObjs_with_peer_end;
	map<Udm::Object, pair<string, string> > newAssocObjs_with_rolenames;
	map<Udm::Object, pair<Udm::Object, Udm::Object> >::iterator ai;
	for(ai=assoc_with_peer_end.begin();ai!=assoc_with_peer_end.end();++ai)
	{
		Udm::Object assocObj = (*ai).first;
		Uml::Class childType = assocObj.type();
		Udm::Object::CompositionInfo compInfo;
		Udm::Object cObj = moveDst.CreateObject(childType, compInfo);
		backendMap->CheckNewObject(cObj);

		UdmUtil::copy_assoc_map dummy;
	//	UdmUtil::CopyOpts opts = { false, false };
		UdmUtil::CopyObjectHierarchy(assocObj.__impl(), cObj.__impl(), cObj.__impl()->__getdn(), dummy);
		newAssocObjs_with_peer_end[cObj] = (*ai).second;
		newAssocObjs_with_rolenames[cObj] = assoc_with_rolenames[assocObj];

		//Check whether any set has this association object as its member, if yes, update it
		for(map<Udm::Object, set<Udm::Object> >::iterator ssi=set_with_members.begin();ssi!=set_with_members.end();++ssi)
		{
			set<Udm::Object> setmembers = (*ssi).second;
			if(setmembers.find(assocObj)!=setmembers.end())
			{
				setmembers.erase(assocObj);
				setmembers.insert(cObj);
			}
		}

		backendMap->DeleteUdmObject(assocObj);
	}

	//Move other udm objects, if it is the Set, move it later
	for(set<Udm::Object>::iterator oi=groupObjs.begin(); oi!=groupObjs.end();++oi)
	{
		Udm::Object mobj = *oi;
		if(set_with_members.find(mobj) == set_with_members.end())
			mobj.__impl()->setParent(moveDst.__impl(), NULL);
	}

	//Move the Set object if there is any and reset the set-member relationship
	for(map<Udm::Object, set<Udm::Object> >::iterator si=set_with_members.begin();si!=set_with_members.end();++si)
	{
		Udm::Object setObj = (*si).first;
		setObj.__impl()->setParent(moveDst.__impl(), NULL);

		set<Udm::Object> members = (*si).second;
		pair<string, string> rolenames = set_with_rolenames[setObj];
		const Uml::Class clsNull(NULL);
		Udm::Object::AssociationInfo set_assoInfo(clsNull);
		set_assoInfo.strSrcRoleName = rolenames.first;
		set_assoInfo.strDstRoleName = rolenames.second;
		for(set<Udm::Object>::iterator mi=members.begin();mi!=members.end();++mi)
		{
			Udm::Object member = *mi;
			member.DeleteLink(setObj, set_assoInfo); //This additional line is for XML and memory backend
			member.CreateLink(setObj, set_assoInfo);
		}
	}

	//reset the association
	for(ai=newAssocObjs_with_peer_end.begin();ai!=newAssocObjs_with_peer_end.end();++ai)
	{
		Udm::Object assocObj = (*ai).first;
		Udm::Object srcObj = ((*ai).second).first;
		Udm::Object dstObj = ((*ai).second).second;

		pair<string, string> rolenames = newAssocObjs_with_rolenames[assocObj];
		const Uml::Class clsNull(NULL);
		Udm::Object::AssociationInfo src_assoInfo(clsNull);
		src_assoInfo.strSrcRoleName = rolenames.first;
		src_assoInfo.strDstRoleName = rolenames.second;

		Udm::Object::AssociationInfo dst_assoInfo(clsNull);
		dst_assoInfo.strSrcRoleName = rolenames.second;
		dst_assoInfo.strDstRoleName = rolenames.first;

		srcObj.CreateLink(assocObj, src_assoInfo);
		dstObj.CreateLink(assocObj, dst_assoInfo);
	}
}

void GroupEffector::DeleteGroup(vector<Match> &group, set<GR::ObjectLink> &groupLinks, 
								set<Udm::Object> &groupAssocs, set<Udm::Object> &groupObjs)
{
	set<Udm::Object>::iterator si;
	for(set<GR::ObjectLink>::iterator li=groupLinks.begin();li!=groupLinks.end();++li)
	{
		GR::ObjectLink grLink = *li;
		GR::ObjectWrapper grSrc = grLink.src_end();
		GR::ObjectWrapper grDst = grLink.dst_end();

		const Uml::Class clsNull(NULL);
		Udm::Object::AssociationInfo assoInfo(clsNull);
		assoInfo.strSrcRoleName = grLink.srcRoleName();
		assoInfo.strDstRoleName = grLink.dstRoleName();
	
		for(vector<Match>::iterator mi=group.begin();mi!=group.end();++mi)
		{
			Match match = *mi;
			(match.grObj2BoundObj[grSrc]).DeleteLink(match.grObj2BoundObj[grDst], assoInfo);
		}
	}
	for(si=groupAssocs.begin();si!=groupAssocs.end();++si)
		backendMap->DeleteUdmObject(Udm::Object(*si));
	for(si=groupObjs.begin();si!=groupObjs.end();++si)
		backendMap->DeleteUdmObject(Udm::Object(*si));
}

void GroupEffector::reqDelete(map<GR::ObjectWrapper, list<Udm::Object> > &boundObjs)
{
	set<GR::GRBase> groupMembers = _group.members();  //get group members

	set<GR::ObjectLink> grLinks = boundRule->ObjectLink_kind_children();
	for(set<GR::ObjectLink>::iterator li=grLinks.begin();li!=grLinks.end();++li)
	{
		GR::ObjectLink grLink = *li;
		if((string)grLink.action() != "Delete") continue;
		if((string)grLink.relationshipType() != "SimpleAssociation") continue;
		if(groupMembers.find(grLink)!=groupMembers.end()) continue;
		
		GR::ObjectWrapper grSrc = grLink.src_end();
		GR::ObjectWrapper grDst = grLink.dst_end();

		const Uml::Class clsNull(NULL);
		Udm::Object::AssociationInfo assoInfo(clsNull);
		assoInfo.strSrcRoleName = grLink.srcRoleName();
		assoInfo.strDstRoleName = grLink.dstRoleName();

		list<Udm::Object> srcObjs = boundObjs[grSrc];
		list<Udm::Object> dstObjs = boundObjs[grDst];
		list<Udm::Object>::iterator si=srcObjs.begin();
		list<Udm::Object>::iterator di=dstObjs.begin();
		while(si!=srcObjs.end() && di!=dstObjs.end())
		{
			Udm::Object srcObj = *si;
			Udm::Object dstObj = *di;
			srcObj.DeleteLink(dstObj, assoInfo);
			si++;
			di++;
		}
	}

	set<GR::ObjectWrapper> grObjs = boundRule->ObjectWrapper_kind_children();
	set<GR::ObjectWrapper>::iterator oi;
	for(oi=grObjs.begin();oi!=grObjs.end();++oi)
	{
		GR::ObjectWrapper grObj = *oi;
		if((string)grObj.action()!="Delete") continue;
		if(!grObj.isAssociationClass()) continue;
		if(groupMembers.find(grObj)!=groupMembers.end()) continue;
		list<Udm::Object> assoc_listObjs = boundObjs[grObj];
		set<Udm::Object> assoc_setObjs;
		list2set(assoc_listObjs, assoc_setObjs);
		for(set<Udm::Object>::iterator ai=assoc_setObjs.begin();ai!=assoc_setObjs.end();++ai)
			backendMap->DeleteUdmObject(Udm::Object(*ai));
	}

	for(oi=grObjs.begin();oi!=grObjs.end();++oi)
	{
		GR::ObjectWrapper grObj = *oi;
		if((string)grObj.action()!="Delete") continue;
		if(grObj.isAssociationClass() || groupMembers.find(grObj)!=groupMembers.end()) continue;
		list<Udm::Object> listObjs = boundObjs[grObj];
		set<Udm::Object> setObjs;
		list2set(listObjs, setObjs);
		for(set<Udm::Object>::iterator ii=setObjs.begin();ii!=setObjs.end();++ii)
			backendMap->DeleteUdmObject(Udm::Object(*ii));
	}
}

void GroupEffector::list2set(list<Udm::Object> &listObjs, set<Udm::Object> &setObjs)
{
	for(list<Udm::Object>::iterator i=listObjs.begin();i!=listObjs.end();++i)
	{
		setObjs.insert(*i);
	}
}
typedef bool (*MAP_GROUPFN)(const vector <Udm::Object> &the_objs, const vector <Udm::Object> &other_objs, bool& err);
bool GroupEffector::executeGroupRule(Match &the_match, Match &other_match)
{
	string groupRule = _group.rule();
	if(groupRule.empty()) return true;

	ostringstream gFStream;
	const __int64 &o_id = boundRule->ruleId();
	gFStream << "Group_" << convertToCPlusPlusName((string)boundRule->name()) << "_" << o_id ;
	string fn = gFStream.str();

	MAP_GROUPFN lpMapFn = (MAP_GROUPFN)GetProcAddress(hCINTDLL, fn.c_str());

	if(lpMapFn == NULL)
	{	cout << "ERROR: Group function " << fn << " not found in dll \n";
		return true;
	}

	vector <Udm::Object > the_objs, other_objs;
	set<string > names;
	string name;

	typedef NameCompareFunctor< GR::PatternObject> PatternObjectCompareFunc;
	typedef set< GR::PatternObject, PatternObjectCompareFunc> PattObjSorted_t;
	PattObjSorted_t patobjs = boundRule->PatternObject_kind_children_sorted( PatternObjectCompareFunc());
	PattObjSorted_t::const_iterator  i;

	for(i=patobjs.begin();i!=patobjs.end();i++)
	{
		const GR::PatternObject &po = *i;
		if((string)po.action() == "CreateNew") continue;

		name = po.name();
		set<string>::iterator ni = names.find(name);
		if(ni!=names.end())
			continue;
		names.insert(name);

		Udm::Object o;
		if(the_match.grObj2BoundObj.find(po)!=the_match.grObj2BoundObj.end())
		{
			o = the_match.grObj2BoundObj[po];
		}
		the_objs.push_back(o);
		if(other_match.grObj2BoundObj.find(po)!=other_match.grObj2BoundObj.end())
		{
			o = other_match.grObj2BoundObj[po];
		}
		other_objs.push_back(o);
	}


	bool err = false;
	bool ret = false;
	ret = lpMapFn(the_objs,other_objs,err);
	if (err)
    cout << "ERROR: Gorup function " << fn << " could not execute\n";

	return ret;
}

typedef void (*MAP_FN)(const vector <list<Udm::Object> > &objs, bool& err);
void GroupEffector::executeCode(map<GR::ObjectWrapper, list<Udm::Object> > &boundObjs)
{	
	const vector<GR::Code> &ams = boundRule->Code_kind_children();
	if(ams.size()==0)
		return;

	ostringstream gFStream;
	const __int64 &o_id = boundRule->ruleId();
	gFStream << "AM" << "_" << convertToCPlusPlusName((string)boundRule->name()) << "_" << o_id ;
	string fn = gFStream.str();

	//cout << "Executing AM function " << fn << endl;

  MAP_FN lpMapFn = (MAP_FN)GetProcAddress(hCINTDLL, fn.c_str());

	if(lpMapFn == NULL)
	{	
    cout << "ERROR: AM function " << fn << " not found in dll \n";
		return;
	}

	vector <list<Udm::Object> > objs;
	set<string > names;

	typedef NameCompareFunctor< GR::PatternObject> PatternObjectCompareFunc;
	typedef set< GR::PatternObject, PatternObjectCompareFunc> PattObjSorted_t;
	PattObjSorted_t patobjs = boundRule->PatternObject_kind_children_sorted( PatternObjectCompareFunc());
	PattObjSorted_t::const_iterator i;

	for(i=patobjs.begin();i!=patobjs.end();i++)
	{	const GR::PatternObject &po = *i;
		string name = po.name();
		set<string>::iterator ni = names.find(name);
		if(ni!=names.end())
			continue;
		names.insert(name);
		unsigned long o_id = (unsigned long)po.objectID();
		list<Udm::Object> o;
		if(boundObjs.find(po)!=boundObjs.end())
		{
			o = boundObjs[po];
			objs.push_back(o);
		}
	}

  bool err = false;
  lpMapFn(objs, err);
  if (err)
    cout << "ERROR: AM function " << fn << " could not execute \n";
	return;
}

void GroupEffector::createRTT(map<GR::ObjectWrapper, list<Udm::Object> > &boundObjs)
{
	const set<GR::ObjectLink>& grLinks = boundRule->ObjectLink_kind_children();
	for(set<GR::ObjectLink>::const_iterator i=grLinks.begin();i!=grLinks.end();++i)
	{
		const GR::ObjectLink& plink = *i;
		if((string)plink.relationshipType()!="TracesTo") continue;
		
		GR::ObjectWrapper srcGRObj = plink.src_end();
		GR::ObjectWrapper dstGRObj = plink.dst_end();

		Udm::Object srcObj, dstObj;
		if(boundObjs.find(srcGRObj)!=boundObjs.end() && boundObjs.find(dstGRObj)!=boundObjs.end()    )
		{
			list<Udm::Object> srcObjs = boundObjs[srcGRObj];
			list<Udm::Object> dstObjs = boundObjs[dstGRObj];
			
			for(list<Udm::Object>::iterator sit=srcObjs.begin();sit!=srcObjs.end();++sit)
			{
				Udm::Object srcObj = *sit;
				for(list<Udm::Object>::iterator dit=dstObjs.begin();dit!=dstObjs.end();++dit)
				{
					Udm::Object dstObj = *dit;
					RTTGenerator::Instance()->generateRTT(srcObj.uniqueId(), srcObj.__impl()->__getdn()->uniqueId(), UdmUtil::ExtractName(srcObj),
														  dstObj.uniqueId(), dstObj.__impl()->__getdn()->uniqueId(), UdmUtil::ExtractName(dstObj));
				}
			}
		}
	}
}