/* NAME:	Effector
 * AUTHOR:	Attila Vizhanyo - viza@isis-server.isis.vanderbilt.edu
 *
 * Generate code, which executes actions 'create new' and 'delete' on 
 * 'GR::PatternObject'-s and 'GR::ObjectLink'-s.
 */
#ifndef __EFFECTOR_H
#define	__EFFECTOR_H

#ifdef _MSC_VER
#pragma warning (disable : 4786) 
#pragma warning (disable : 4503)
#endif

#include "RuleTypes.h"
#include "PE.h"
#include "GeneralBackend.h"
#include "PacketPassing.h"
#include "ICIntCompiler.h"

class CodeForDeleteClassFunctor; // forward declaration

// Functor which creates code for traversing & deleting association classes. Used by 'CodeForDeleteClassFunctor'.
class CodeForAssocRoleFunctor : public std::unary_function< Uml::AssociationRole, void>
{
public:
	// ctor; takes bound objects, delete class functor, the pattern object of which associations is being traversed, and the output code. 
	CodeForAssocRoleFunctor( std::map< GR::ObjectWrapper, std::string>& boundObjects, const CodeForDeleteClassFunctor& deleteClassFunctor, const GR::PatternObject& pattern, PEComposite* output) : 
	  _boundObjects( boundObjects), _deleteClassFunctor( deleteClassFunctor), _pattern( pattern), _output( output) 
	  {}
	// func.op.; takes association role 'role',
	// if 'role' is kind of association role, delete code is created for associations identified by 'role.name()'.
	void operator()( const Uml::AssociationRole& role) const;

private:
	std::map< GR::ObjectWrapper, std::string>& _boundObjects;
	PEComposite* _output;
	const CodeForDeleteClassFunctor& _deleteClassFunctor;
	const GR::PatternObject& _pattern;
};

// Functor which creates code for deleting classes.
class CodeForDeleteClassFunctor : public std::binary_function< GR::PatternObject, bool, bool>
{
public:
	// ctor; takes bound objects, host meta backend, and the output code. 
	CodeForDeleteClassFunctor( std::map< GR::ObjectWrapper, std::string>& boundObjects, const GeneralBackend& backend, PEComposite* output) : 
	  _boundObjects( boundObjects), _backend( backend), _output( output) 
	  {}
	// func.op.; creates code which deletes the class referred by 'pattern', 
	// only if 'isAssocClass== pattern.isAssociationClass()'.
	bool operator()( const GR::PatternObject& pattern, bool isAssocClass) const;

protected:
	// Generate code to 'addTo', which performs object deletion on 'name'.
	void deleteByName( const std::string& name, PEComposite* addTo) const;
	// Traverses associations of class referred by 'pattern', 
	// and generate association class deletion code for association class type associations.
	void deletePeerAssocClassObjects( const GR::PatternObject& pattern) const;

private:
	friend class CodeForAssocRoleFunctor; // provide access to deleteByName()
	std::map< GR::ObjectWrapper, std::string>& _boundObjects;
	PEComposite* _output;
	const GeneralBackend& _backend;
};

// Effector code generation. 
// Those object wrappers with GR::GRBase::action()= "CreateNew" or "Delete".
// This class is not designed to be a base class. 
// Either don't inherit from it, or add virtual destructor.
class Effector
{
public:
	// ctor; see parameter desc. at the member variables declaration part. 
	Effector( const GRSimpleRule& rule, 
		const GeneralBackend& backend, 
		std::map< GR::ObjectWrapper, std::string>& boundObjects,
		ICIntCompiler& cc
		) : _rule( rule) , _backend( backend), _boundObjects( boundObjects), _currCode( NULL), _cc( cc) 
		{}
	// Generate effector code into '_currCode'. 
	// (Name was preversed from GReAT to have similar interfaces)
	virtual void executeRightSide( PEComposite& code);

	// Gain access to the generated code.
	const PEComposite* getCode() const { return _currCode;}
	PEComposite* getCode() { return _currCode;}

protected:	
	// Create code which creates objects.
	void createObjects();
	// Create code which create/ delete object links.
	// (Name was preversed from GReAT to have similar interfaces)
	virtual void bindRemainingObjectLinks();
	// Returns with a 'parentOW'-'childOW' pair where 'parentOW' action is bound and 'childOW' action is 'CreateNew'. 
	// Return value is true, if such pair found, false otherwise.
	bool getCompositionToCreate( GR::ObjectWrapper& parentOW, GR::ObjectWrapper& childOW, std::string& childRole) const;
	// Create code which creates new object referred by 'childOW', with parent object reffered by 'parentOW'.
	virtual void createOutputObject( const GR::ObjectWrapper& parentOW, GR::ObjectWrapper& childOW, const std::string& childRole);
	// Create code which delete association class type objects.
	void deleteAssocClassObjects();
	// Create code which delete objects. 
	// The difference between 'deleteAssocClassObjects()' and 'deleteObjects()' is that
	// the later will also delete all association classes associated with the deleted object.
	void deleteObjects();
	// Create code which creates various type object links.
	virtual void createNewObjectLink( const GR::ObjectLink& linkToCreate);
	// Create code which deletes various type object links.
	void deleteObjectLink( const GR::ObjectLink& linkToDelete);
	// Create code which iterates matches.
	void createMatchLoop();
	// Returns with the maximum cardinality of association 'association' linked with class 'boundclass' in paradigm 'paradigm'. 
	// Caller must set 'isSourceAssoc', if 'boundclass' is an association class. 
	int getAssociationCardinality( const std::string& paradigm, const std::string& boundclass, const std::string& association, bool isSourceAssoc) const;
	int getAssociationCardinality( const std::string& paradigm, const std::string& ns_name, const std::string& boundclass, const std::string& association, bool isSourceAssoc) const;
	// Create code for RTT
	void createRTT();

protected:
	// Rule, on which effector code generation executes.
	GRSimpleRule _rule;
	// pointer to program element, which is the current target of pattern matching code generation.
	PEComposite* _currCode;
	// Maps object wrappers to their generated variable names. 
	// Newly created objects get registered by effector.
	// This data structure is also updated by pattern matcher.
	std::map< GR::ObjectWrapper, std::string>& _boundObjects;
	// Host meta backend.
	const GeneralBackend& _backend;
	//
	ICIntCompiler& _cc;
};
#endif //__EFFECTOR_H