/* NAME:	Packet passing
 * AUTHOR:	Attila Vizhanyo - viza@isis-server.isis.vanderbilt.edu
 *
 * Tracks and matches packets between rules.
 */
#ifndef __PACKETPASSING_H
#define	__PACKETPASSING_H

#ifdef _MSC_VER
#pragma warning (disable : 4786)
#pragma warning (disable : 4503) 
#endif

#include <UmlExt.h>
#include "GR.h"
#include "PE.h"
#include "CGEx.h"
#include "Logging.h"

///////////////////////////////////////////////////////////////////////////////
// Match an object wrapper, which is input or output of a rule, 
// to the parameter name used for that input/ output in the function signature.
class ParamData
{
public:
	// ctor & dtor
	ParamData( const GR::ObjectWrapper& param, const std::string& name) : _param( param), _name( name), _memberDef( NULL) {}
	virtual ~ParamData() {}
	// Conversion op.
	operator const GR::ObjectWrapper() const { return _param;}
	// Input/ output object wrapper.
	GR::ObjectWrapper _param;
	// Matching generated parameter name.
	std::string _name;
	// Input / Output class member definition. (optional)
	const ClassDef* _memberDef;
};

typedef std::list< ParamData> ParamDatas_t;

// Comparision operators. ParamData type objects are compared based on there's '_param' data member.
bool operator<( const ParamData& first, const ParamData& second);
bool operator==( const ParamData& first, const ParamData& second);

///////////////////////////////////////////////////////////////////////////////
// ParamList represents the complete parameter list of a function
// generated for a given rule.
class ParamList
{
public:
	// ctor & dtor
	ParamList() : _function( NULL) {}
	ParamList( Function& function) : _function( &function) {}
	virtual ~ParamList() {}
	// Func.op. adds 'ow' to '_inputParams' or '_outputParams' whether 'ow' is an input or output.
	void operator () ( const GR::ObjectWrapper& ow);	
	// Parameter std::list accessors.
	const ParamDatas_t* inputParams() const { return &_inputParams;}
	const ParamDatas_t* outputParams() const { return &_outputParams;}
	ParamDatas_t* inputParams() { return &_inputParams;}
	ParamDatas_t* outputParams() { return &_outputParams;}

protected:
	// Add parameters to signature of function: '_function'.
	// Return with the generated parameter name.
	const std::string& addInputParam( const GR::ObjectWrapper& ow, Function& function) const;
	const std::string& addOutputParam( const GR::ObjectWrapper& ow, Function& function) const;

private:
	// Func. signature to append with parameters by calling func.op.
	Function* _function;	// not-owned pointer
	// List of inputs.
	ParamDatas_t _inputParams;
	// List of outputs.
	ParamDatas_t _outputParams;
};

// Input / Output port maintenance.
void registerPort( const GR::ObjectWrapper& port, const ClassDef* memberDef, ParamDatas_t& params) throw ( CGEx);
// Returns NULL if not found.
const ClassDef* getPort( const GR::ObjectWrapper& port, const ParamDatas_t& params);


///////////////////////////////////////////////////////////////////////////////
class DeletedPredicate : public std::unary_function< GR::ObjectWrapper, bool>
{
public:
	// Returns true, if 'ow' is an object to delete
	bool operator()( const GR::ObjectWrapper& ow) const
	{
		return ( std::string) ow.action() == "Delete";
	}
};
// More comfortable name for the predicate.
typedef DeletedPredicate isDeleted;

///////////////////////////////////////////////////////////////////////////////
class InputPredicate : public std::unary_function< GR::ObjectWrapper, bool>
{
public:
	// Returns true, if 'ow' is an input of a rule.
	bool operator()( const GR::ObjectWrapper& ow) const
	{
		// Inputs are 'InputPort'-s and 'PatternObject'-s having at least one PassAlong source.
		if ( Uml::IsDerivedFrom( ow.type(), GR::OutputPort::meta))
			return false;
		if ( Uml::IsDerivedFrom( ow.type(), GR::InputPort::meta))
			return true;
		std::set< GR::PassAlong> srcOW= ow.srcObjectWrapper();
		if ( 0< srcOW.size())
			return true;
		return false;
	}
};
// More comfortable name for the predicate.
typedef InputPredicate isInput;

///////////////////////////////////////////////////////////////////////////////
class OutputPredicate : public std::unary_function< GR::ObjectWrapper, bool>
{
public:
	// Returns true, if 'ow' is an output of a rule.
	bool operator()( const GR::ObjectWrapper& ow) const
	{
		// Outputs are 'OutputPort'-s and 'PatternObject'-s having at least one PassAlong destination.
		if ( Uml::IsDerivedFrom( ow.type(), GR::InputPort::meta))
			return false;
		if ( Uml::IsDerivedFrom( ow.type(), GR::OutputPort::meta))
			return true;
		std::set< GR::PassAlong> dstOW= ow.dstObjectWrapper();
		if ( 0< dstOW.size())
			return true;
		return false;
	}
};
// More comfortable name for the predicate.
typedef OutputPredicate isOutput;

///////////////////////////////////////////////////////////////////////////////
// General Packet Passing Interface. 
// Bunch of various functions used for packet passing.
// This class is not designed to be a base class. 
// Either don't inherit from it, or add virtual destructor.
class PacketPassing
{
public:
	static bool isAnyDeletedObjectAmongParams( const ParamDatas_t& params);
	static bool allOutputsAreInputs( const ParamDatas_t& outputs);
	// Fills parameter list of function: 'function', with inputs & outputs of 'rule'.
	void createParamList( const GR::RewritingRule& rule, Function& function);
	// Adds input iterator for 'inputParam' into loop header of loop: 'inputLoop'.
	const std::string& addInputIterator( const ParamData& inputParam, Loop& inputLoop) const;
	// Iterating inputs of 'destination', the matching source parameter names of 'source' are added to 'callerParams'.
	void matchParams( const GR::RuleBase& source, const GR::RuleBase& destination, std::list< std::string>& callerParams, bool findSourceInInputs) const;
	// Returns with inputs/ outputs of 'rule'. Returns NULL if not found.
	const ParamDatas_t* inputParams( const GR::RewritingRule& rule) const;
	const ParamDatas_t* outputParams( const GR::RewritingRule& rule) const;
	ParamDatas_t* inputParams( const GR::RewritingRule& rule);
	ParamDatas_t* outputParams( const GR::RewritingRule& rule);

protected:
	// Get input object wrapper for pass along tracking. 
	// If 'ruleBase' is of type 'GR::RewritingRuleProxy', return with object wrapper proxy, which refers to 'ow'.
	// Return 'ow', otherwise.
	GR::ObjectWrapper getInputOWForPATracking( const GR::RuleBase& ruleBase, const GR::ObjectWrapper ow) const;
	// Find parameter data in 'params' with key
	// 'owToFind.refersTo()' if 'owToFind' of type 'GR::ObjectWrapperProxy'
	// 'owToFind', otherwise.
	const ParamData& findParamData( const ParamDatas_t& params, const GR::ObjectWrapper& owToFind) const;

private:
	// Maps rules to their parameters.
	typedef std::map< GR::RewritingRule, ParamList> RuleParamMap_t;
	RuleParamMap_t _ruleParams;
};

#endif //__PACKETPASSING_H