/* NAME:	RuleTypes
 * AUTHOR:	Attila Vizhanyo - viza@isis-server.isis.vanderbilt.edu
 *
 * Strongly typed classes for different types of GR::RewritingRule. 
 * Aim was to get run-time exceptions in case of type mismatches.
 * There is no way to get a compile-time error if one passes a wrong type 
 * GR::RewritingRule, beacuse type is inherently represented as a set of
 * class data members. But it's strongly desired to get an exception when
 * illegal type mismatch occurs.
 */
#ifndef __RULETYPES_H
#define __RULETYPES_H

#ifdef _MSC_VER
#pragma warning (disable : 4786)
#pragma warning (disable : 4503) 
#endif

#include <string>
#include "GR.h"
#include "CGEx.h"

// Proxy resolver.
// Returns 'ruleBase.refersTo()', if 'ruleBase' is of type GR::RewritingRuleProxy.
// Returns 'ruleBase', otherwise.
extern GR::RewritingRule getRewritingRule( const GR::RuleBase& ruleBase);

class GRBlockType : public GR::RewritingRule
{
public:
	GRBlockType( const GR::RewritingRule& from) : GR::RewritingRule( from)
	{
		if ( std::string( "BLOCK") != ( std::string)from.ruleType())
			throw CGEx("Trying to translate invalid type RewritingRule to GRBlockType.");
	}
private:
	// no default ctor
	GRBlockType();
};

class GRBlock : public GRBlockType
{
public:
	GRBlock( const GR::RewritingRule& from) : GRBlockType( from)
	{
		if ( ( std::string( "BLOCK") != ( std::string)from.ruleType()) || ( true==from.forBlock()))
			throw CGEx("Trying to translate invalid type RewritingRule to GRBlock.");
	}
private:
	// no default ctor
	GRBlock();
};

class GRForBlock : public GRBlockType
{
public:
	GRForBlock( const GR::RewritingRule& from) : GRBlockType( from)
	{
		if ( ( std::string( "BLOCK") != ( std::string)from.ruleType()) || ( false==from.forBlock()))
			throw CGEx("Trying to translate invalid type RewritingRule to GRForBlock.");
	}
private:
	// no default ctor
	GRForBlock();
};

class GRTestType : public GR::RewritingRule
{
public:
	GRTestType( const GR::RewritingRule& from) : GR::RewritingRule( from)
	{
		if ( std::string( "TEST") != ( std::string)from.ruleType())
			throw CGEx("Trying to translate invalid type RewritingRule to GRTestType.");
	}
private:
	// no default ctor
	GRTestType();
};

class GRTest : public GRTestType
{
public:
	GRTest( const GR::RewritingRule& from) : GRTestType( from)
	{
		if ( ( std::string( "TEST") != ( std::string)from.ruleType()) || ( true==from.forBlock()))
			throw CGEx("Trying to translate invalid type RewritingRule to GRTest.");
	}
private:
	// no default ctor
	GRTest();
};

class GRForTest : public GRTestType
{
public:
	GRForTest( const GR::RewritingRule& from) : GRTestType( from)
	{
		if ( ( std::string( "TEST") != ( std::string)from.ruleType()) || ( false==from.forBlock()))
			throw CGEx("Trying to translate invalid type RewritingRule to GRForTest.");
	}
private:
	// no default ctor
	GRForTest();
};

class GRSimpleRule : public GR::RewritingRule
{
public:
	GRSimpleRule( const GR::RewritingRule& from) : GR::RewritingRule( from)
	{
		if ( ( std::string( "RULE") != ( std::string)from.ruleType()) && ( std::string( "CASE") != ( std::string)from.ruleType())) 
			throw CGEx("Trying to translate invalid type RewritingRule to GRSimpleRule.");
	}
private:
	// no default ctor
	GRSimpleRule();
};

class GRRule : public GRSimpleRule
{
public:
	GRRule( const GR::RewritingRule& from) : GRSimpleRule( from)
	{
		if ( std::string( "RULE") != ( std::string)from.ruleType()) // attribute forBlock is ignisificant
			throw CGEx("Trying to translate invalid type RewritingRule to GRRule.");
	}
private:
	// no default ctor
	GRRule();
};

class GRCase : public GRSimpleRule
{
public:
	GRCase( const GR::RewritingRule& from) : GRSimpleRule( from)
	{
		if ( std::string( "CASE") != ( std::string)from.ruleType()) // attribute forBlock is ignisificant
			throw CGEx("Trying to translate invalid type RewritingRule to GRCase.");
	}
private:
	// no default ctor
	GRCase();
};


#endif //__RULETYPES_H