#include "Sequencer.h"
#include "Logging.h" // temporaly logger

using namespace std;

Sequencer::Sequencer( Translator& translator) : _depth( 0) , _translator( translator)
{	
}

Sequencer::~Sequencer() 
{
}

void Sequencer::execute( const GR::RuleBase& rule)
{
	std::string mn = rule.type().name();
	if( "RewritingRuleProxy" == mn)
		executeProxy( GR::RewritingRuleProxy::Cast( rule));
	else
		executeRewritingRule( GR::RewritingRule::Cast( rule));
}

void Sequencer::executeProxy( const GR::RewritingRuleProxy& proxy)
{
	logRule( REFERENCE, ( std::string)proxy.name(), _depth);
	execute( proxy.refersTo());
}

void Sequencer::executeRewritingRule( const GR::RewritingRule& rule)
{	
	set<GR::PatternGroup> groups = rule.PatternGroup_kind_children();
	if(!groups.empty())
	{
//		throw udm_exception("Code Generator doesn't support Group for Rule: "+(string)rule.name());
//		return;
	}
	std::string rt = rule.ruleType();
	bool isfor = rule.forBlock();
	if( rt == "BLOCK")
		if( isfor) 
		{
			executeForBlock( rule);
			return;
		}
		else
		{
			executeBlock( rule);
			return;
		}
	if( rt == "TEST")
		if( isfor) 
		{
			executeForTest( rule);
			return;
		}
		else
		{
			executeTest( rule);
			return;
		}
	if( rt == "RULE")
	{
		executeRule( rule);
		return;
	}
	if( rt == "CASE")
	{
		executeCase( rule);
		return;
	}
	printError( "Invalid rule type: "+ rt);
}

void Sequencer::executeBlock( const GRBlock& rule)
{	
	if ( isRuleTraversed( rule))
		return;
	setRuleTraversed( rule);
	logRule( BLOCK, ( std::string)rule.name(), _depth);
	++_depth;

	_translator.translateBlock( rule);
	
	// It is unnecessary to follow the sequence of the consisted rules.
	// Simply generate class definitions for each consisted rules, then later
	// traverse the sequences and generate calls based on that sequence order...
	std::set< GR::RuleBase> ruleBases= rule.RuleBase_kind_children();
	for( std::set< GR::RuleBase>::const_iterator i= ruleBases.begin(); i!= ruleBases.end(); ++i)
		execute( *i);

//	std::stack< GR::RuleBase> dstRules;
//	destinationRules( rule, rule, dstRules);
//	// process destination rules 'nextRule'-s
//	while( !dstRules.empty())
//	{
//		GR::RuleBase dstRule = dstRules.top();
//		dstRules.pop();
//		execute( dstRule);
//		destinationRules( dstRule, rule, dstRules);	// destination rules of 'dstRule' 
//	}

	translateAllCalls( rule);
	--_depth;
}

void Sequencer::executeForBlock( const GRForBlock& rule)
{
	if ( isRuleTraversed( rule))
		return;
	setRuleTraversed( rule);
	logRule( FORBLOCK, ( std::string)rule.name(), _depth);
	++_depth;

	_translator.translateForBlock( rule);

	// It is unnecessary to follow the sequence of the consisted rules.
	// Simply generate class definitions for each consisted rules, then later
	// traverse the sequences and generate calls based on that sequence order...
	std::set< GR::RuleBase> ruleBases= rule.RuleBase_kind_children();
	for( std::set< GR::RuleBase>::const_iterator i= ruleBases.begin(); i!= ruleBases.end(); ++i)
		execute( *i);

//	std::stack< GR::RuleBase> dstRules;
//	destinationRules( rule, rule, dstRules);
//	// process destination rules 'nextRule'-s
//	while( !dstRules.empty())
//	{
//		GR::RuleBase dstRule = dstRules.top();
//		dstRules.pop();
//		execute( dstRule);
//		destinationRules( dstRule, rule, dstRules);	// destination rules of 'dstRule' 
//	}

	translateAllCalls( rule);
	--_depth;
}

void Sequencer::executeTest( const GRTest& rule)
{	
	if ( isRuleTraversed( rule))
		return;
	setRuleTraversed( rule);
	logRule( TEST, ( std::string)rule.name(), _depth);
	++_depth;

	_translator.translateTest( rule);

	// here the order of rules in sequence is important.
	std::stack< GR::RuleBase> dstRules;
	destinationRules( rule, rule, dstRules);
	while( !dstRules.empty())
	{
		GR::RuleBase dstRule = dstRules.top();
		dstRules.pop();
		execute( dstRule);
		
		// 'dstRule' must be of type case! ('test' can contain only 'case'-s!)
		_translator.translateCallInTest( rule, GRCase( GR::RewritingRule::Cast( dstRule)));	

		std::set< GR::Sequence> seqs = dstRule.dstRule();
		if( seqs.size() != 1)
		{	
			printError( "case has more that one destination rule.");
			break;
		}
		GR::RewritingRule ruleAfter = GR::RewritingRule::Cast( seqs.begin()->dstRule_end());
		std::string ruleAfterType = ruleAfter.ruleType();
		if( ruleAfterType== "CASE")
			dstRules.push( ruleAfter);
		else
			break;
	}
	--_depth;
}

void Sequencer::executeForTest( const GRForTest& rule)
{
	if ( isRuleTraversed( rule))
		return;
	setRuleTraversed( rule);
	logRule( FORTEST, ( std::string)rule.name(), _depth);
	++_depth;

	_translator.translateForTest( rule);

	// here the order of rules in sequence is important.
	std::stack< GR::RuleBase> dstRules;
	destinationRules( rule, rule, dstRules);
	while( !dstRules.empty())
	{
		GR::RuleBase dstRule = dstRules.top();
		dstRules.pop();
		execute( dstRule);
		
		// 'dstRule' must be of type case! ('test' can contain only 'case'-s!)
		_translator.translateCallInTest( rule, GRCase( GR::RewritingRule::Cast( dstRule)));	

		std::set< GR::Sequence> seqs = dstRule.dstRule();
		if( seqs.size() != 1)
		{	
			printError( "case has more that one destination rule.");
			break;
		}
		GR::RewritingRule ruleAfter = GR::RewritingRule::Cast( seqs.begin()->dstRule_end());
		std::string ruleAfterType = ruleAfter.ruleType();
		if( ruleAfterType== "CASE")
			dstRules.push( ruleAfter);
		else
			break;
	}
	--_depth;
}

void Sequencer::executeRule( const GRRule& rule)
{	
	if ( isRuleTraversed( rule))
		return;
	setRuleTraversed( rule);
	logRule( RULE, ( std::string)rule.name(), _depth);
	_translator.translateRule( rule);
}

void Sequencer::executeCase( const GRCase& rule)
{
	if ( isRuleTraversed( rule))
		return;
	setRuleTraversed( rule);
	logRule( CASE, ( std::string)rule.name(), _depth);

	_translator.translateCase( rule);
}

void Sequencer::destinationRules( const GR::RuleBase& rule, const GR::RuleBase& parent, std::stack< GR::RuleBase>& dstRules) const
{
	std::set< GR::Sequence> seqs = rule.dstRule();
	for( std::set<GR::Sequence>::const_iterator itSeq= seqs.begin(); itSeq!=seqs.end(); ++itSeq)
	{	
		const GR::Sequence& seq = *itSeq;
		GR::RuleBase dstRule = seq.dstRule_end();
		// if parent of sequence 'seq' is not 'rule'
		if( GR::RewritingRule::Cast( seq.RewritingRule_parent()) != parent)
			continue;
		// if sequence has no destination (possible?)
		if( dstRule == Udm::null)	  
			continue;
		// if sequence connects the block with its parent: passing packets to output port.
		if( dstRule == parent) 	  
			continue;
		// if that rule has already been traversed
		GR::RewritingRule RRule= getRewritingRule( dstRule);
//		std::cout << ( std::string) rule.name() << "->" << ( std::string) RRule.name() << std::endl;
		if ( RRule!= dstRule)
			dstRules.push( dstRule);	// proxy
		else
		{
			if ( true == isRuleTraversed( RRule))
			{
				if ( true== isSource( RRule, rule))
				{
					printLogMsg("INFO: Loop found: "+ ( std::string)rule.name()+ "->"+ ( std::string)dstRule.name());
					continue;
				}
			}
			dstRules.push( dstRule);	// put rule to std::stack
		}
	}
}

// this function is called resursively.
bool Sequencer::isSource( const GR::RuleBase& src, const GR::RuleBase& dst) const
{
	std::string srcname= src.name();
	std::string dstname= dst.name();
	// Is there a path from 'dst' to 'src'?
	std::set <GR::Sequence> srcSeqs= dst.srcRule();
	// First try to find in the source rules of 'dst'.
	for( std::set< GR::Sequence>::const_iterator itSrcSeqs= srcSeqs.begin(); itSrcSeqs!= srcSeqs.end(); ++itSrcSeqs)
	{
		GR::RuleBase srcRuleSrcSeq= itSrcSeqs->srcRule_end();
		if ( srcRuleSrcSeq== src)
			return true;
	}
	// If not found, treat sources as 'dst', and call 'isSource' on each, until found.
	for( std::set< GR::Sequence>::const_iterator itSrcSeqs2= srcSeqs.begin(); itSrcSeqs2!= srcSeqs.end(); ++itSrcSeqs2)
	{
		if ( GR::RewritingRule::Cast( itSrcSeqs2->parent()) != GR::RewritingRule::Cast( src.parent()))
			break;
		GR::RuleBase srcRuleSrcSeq= itSrcSeqs2->srcRule_end();
		if ( true== isSource( src, srcRuleSrcSeq))
			return true;
	}
	// not found, 'src' is not source of 'dst'.
	return false;
}

void Sequencer::translateAllCalls( const GRBlockType& block)
{
	std::set< GR::Sequence> callerSeqs= block.Sequence_kind_children();
	// Collect input sequences.
	std::list< GR::Sequence> inputSeqs;
	for( std::set< GR::Sequence>::const_iterator itSeqs= callerSeqs.begin(); itSeqs!= callerSeqs.end(); ++itSeqs)
	{
		if ( block== GR::RuleBase::Cast( itSeqs->srcRule_end()))
			inputSeqs.push_back( *itSeqs);
	}
	// Start traverse with the input sequences.
	traverseSequences( inputSeqs);
}

// traverseSequences traverses recursively the sequences 'currSeqs'.
// A std::set of sequences of a rule can be viewed as a directed (possibly cyclic) graph, where 
// - edges are the sequences, and 
// - vertices are the end points of sequences (rewriting rules).
// traverseSequences performs an in-breadth traverse on this graph.
// The traversing ends, when 'currSeqs' is empty, thus all sequences traversed.
void Sequencer::traverseSequences( const std::list< GR::Sequence>& currSeqs)
{
	translateSequences( currSeqs);
	std::list< GR::Sequence> destinations;
	for( std::list< GR::Sequence>::const_iterator itSeqs= currSeqs.begin(); itSeqs!= currSeqs.end(); ++itSeqs)
	{
		const GR::Sequence& currSeq= *itSeqs;
		GR::RuleBase currDstRule= currSeq.dstRule_end();
		std::set< GR::Sequence> currDstRuleDestinations= currDstRule.dstRule();
		for( std::set< GR::Sequence>::const_iterator itDstSeqs= currDstRuleDestinations.begin(); itDstSeqs!= currDstRuleDestinations.end(); ++itDstSeqs)
		{
			const GR::Sequence& currDstSeq= *itDstSeqs;
			// Do not cross block boundaries. Traverse only sequences inside block.
			if ( !isSeqTraversed( currDstSeq) && ( GR::RewritingRule::Cast( currDstSeq.parent()) == GR::RewritingRule::Cast( currSeq.parent())))
				destinations.push_back( currDstSeq);
		}
	}
	if ( !destinations.empty())
		traverseSequences( destinations);
}

void Sequencer::translateSequences( const std::list< GR::Sequence>& seqs)
{
	for( std::list< GR::Sequence>::const_iterator itSeqs= seqs.begin(); itSeqs!= seqs.end(); ++itSeqs)
	{
		const GR::Sequence& currSeq= *itSeqs;
		// If sequence has been already traversed...
		if ( isSeqTraversed( currSeq))
			continue;	// no need of function call generation. (never can get here, since 'traverseSequences()' filtered out.
		// Generate the function call.
		_translator.translateCallInBlock( currSeq);
		// Set sequence traversed.
		setSeqTraversed( currSeq);
	}
}

bool Sequencer::isRuleTraversed( const GR::RewritingRule& rule) const
{
	RuleSet_t::const_iterator itFind= find( _traversedRules.begin(), _traversedRules.end(), rule);
	return ( itFind== _traversedRules.end()) ? false : true;
}

void Sequencer::setRuleTraversed( const GR::RewritingRule& rule)
{
	_traversedRules.insert( rule);
}

bool Sequencer::isSeqTraversed( const GR::Sequence& seq) const
{
	TraversedSeqs_t::const_iterator itFind= find( _traversedSeqs.begin(), _traversedSeqs.end(), seq);
	return ( itFind == _traversedSeqs.end()) ? false: true;
}

void Sequencer::setSeqTraversed( const GR::Sequence& seq)
{
	_traversedSeqs.insert( seq);
}
