#ifndef UNIQUEFUNCTOR_H
#define UNIQUEFUNCTOR_H

#include <algorithm>	// for adjacent_find, for_each
#include <vector>
#include <string>

#include <UdmBase.h>

// Not a user type.
class UniqueFunctor {
protected:
	UniqueFunctor() {}	// abstract type, do not instantiate
	virtual ~UniqueFunctor() {}
	template < class T> const T isUnique( T first, T last) {
		std::sort( first, last);
		const T it= std::adjacent_find( first, last);
		return it;
	}
};

// Check of unique attribute values for a sequence of Udm objects.
// Ex.:	std::vector< Udm::Object> objs; 
//		UniqueAttrsFunctor uaf( "attr_name");
//		uaf= for_each( objs.begin(), objs.end(), uaf);
//		if ( false == uaf.isUnique())
//			std::cerr << "Duplicated value found: " << uaf.duplicated() << std::endl;
class UniqueAttrsFunctor : private UniqueFunctor {
public:
	// ctor. 'attr' is the name of the attribute of which values are checked for uniqueness.
	UniqueAttrsFunctor( const std::string& attr) : UniqueFunctor(), _attr( attr) {}
	// Inserts attribute 'attr' of object 'o' into a vector. 
	// Throws 'udm_exception', if 'o' has no attribute 'attr'.
	void operator() ( const Udm::Object& o) {
		std::string val;
		bool res= o.GetStrValue( _attr, val);
		if ( res)
			_vals.push_back( val);
		else
			throw udm_exception("Invalid attribute name: "+ _attr);
	}
	// Returns true is duplicated attribute value found.
	bool isUnique( ) {
		std::vector< std::string>::const_iterator it= UniqueFunctor::isUnique( _vals.begin(), _vals.end());
		return it== _vals.end() ? true : /*false*/ !(*++it == (_dupVal= *it));
	}
	// Returns the duplicated value found by isUnique().
	const std::string& duplicated() const { return _dupVal;}

protected:
	std::string _attr;		// attribute name
	std::string _dupVal;	// duplicated value (if any)
	std::vector< std::string> _vals;	// vector of values
};

// Check of unique attribute values for a sequence of Udm objects.
// Ex1:	Udm::Object obj; 
//		EmptyAttrFunctor eaf( "attr_name");
//		if ( eaf( obj)
//			std::cerr << "Empty value found." << std::endl;
// Ex2: std::vector< Udm::Object> objs;
//		EmptyAttrFunctor eaf( "attr_name");
//		std::vector< Udm::Object>::const_iterator itFindEmpty= std::find_if( objs.begin(), objs.end(), eaf);
//			std::cerr << "The " << (itFindEmpty-objs.begin()) << "th element has empty attribute value." << std::endl;
// (This is an adoptable functor.)
class EmptyAttrFunctor : public std::unary_function< Udm::Object, bool> {
public:
	// ctor. 'attr' is the name of the attribute of which values are checked for emptiness.
	EmptyAttrFunctor( const std::string& attr) : _attr( attr) {}
	// Returns true if the value of attribute 'attr' in object 'o' is empty. Returns false, otherwise.
	// Throws 'udm_exception', if 'o' has no attribute 'attr'.
	bool operator() ( const Udm::Object& o) {
		std::string val;
		bool res= o.GetStrValue( _attr, val);
		if ( res)
			return val.empty();
		else
			throw udm_exception("Invalid attribute name: "+ _attr);
	}

private:
	std::string _attr;		// attribute name
};


#endif //UNIQUEFUNCTOR_H
